<?php if (!defined('PmWiki')) exit();

/*
  UserAuth2 (Version 2.0-beta6) - A user-based page permission granting and authentication module.

  Copyright 2007 Thomas Pitschel, pmwiki@sigproc.de
 
  (Files UserSessionVars.php and userauth2-admintool.php adapted from userauth recipe 
   by James McDuffie et al., see www.pmwiki.org/wiki/Cookbook/UserAuth)

  This is a authentication module for the pmwiki engine that supports user-based
  permission granting via a typical web interface and passwords/cookie based
  authentication. Main features are:

   * definition of user groups
   * ip address based silent permission granting
   * setting login restrictions based on ip address
   * implicit capability of delegating permission granting (via hierarchically structuring
     the users)
   * caching of all permission queries 
   * cookie based authentication

  Detailed instructions on installation and usage can be found at 
  www.pmwiki.org/wiki/Cookbook/UserAuth2

  Bug reports and similar development related suggestions or questions should go to
  www.pmwiki.org/wiki/Cookbook/UserAuth2Devel, unless otherwise said at Cookbook.UserAuth2.

  ThomasP, March 2007.

  Note that if the line after this paragraph does not contain a comment "// COMMENTED VERSION"
  you have got the version of the script where the one-line comments have been stripped. Get a
  debug version then from www.pmwiki.org/wiki/Cookbook/UserAuth2.
  // COMMENTED VERSION

  -- License --

  GNU General Public License. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

*/

define(USERAUTH_VERSION, '2.0');  // compatibility
define(USER_AUTH_VERSION, '2.0'); // compatibility
define(USERAUTH2_VERSION, '2.0');

$RecipeInfo['UserAuth2']['Version'] = '2007-04-02';

/*
  List of the main public functions to be called by application code:

  function UserAuth($pagename, $level, $authprompt=true) // for backward compatibility
  function UserAuth2($pagename, $level, $authprompt=true)
  function TryAccessingPage($pagename, $level)
  function HasCurrentUserPerm($page, $level)
  function HasCurrentUserPermForAction($page, $action) // wrapper employing $HandleAuth to resolve action to level
  function CheckUserPerms($user, $page, $level, $groupaction = false)
  function CheckUserPermsForAction($user, $page, $action, $groupaction = false) // again wrapper of same kind
  function mayCurrAdminActOnPermHolder($action, $admin_action, $tool_username, $groupaction = false)
*/

//==================================================
//======== Default/initial settings ================

SDV($UA2EnforceClientIpFixation, true);     // Ensures that over the course of a session the client ip must not change
SDV($UA2SiteIdentifier,          __FILE__); // If deploying MANY sites on ONE server sharing ONE COMMON session 
                                            // save path and all using THIS auth module, this should be set to different 
                                            // values, otherwise cross-use of authentication status could be possible.
                                            // Using the file location as characteristic should be enough since different 
                                            // pmwikis on one server usually dont share the cookbook dir.
SDV($UA2SessionMaxInactivityTime, 2*60*60); // In seconds, default 2 hours.
SDV($UA2SessionMaxLifeTime,      24*60*60); // In seconds, default 1 day. 
  // Both limits are enforced by this script, no PHP interna are relied upon. However, make sure that
  // session.gc_maxlifetime and session.cache_expire are honoured, see #013, otherwise your session may die earlier.
  // See comments at http://hk2.php.net/manual/en/ref.session.php for details.
SDV($UA2AllowPostCompletion,     true);     // if enabled, posts that hit a session expiration are completed after relogin
                                            // (note this drills a small hole into the session expiration principle)

SDV($UA2EnablePermCaching,       true); // might be useful to set to false during debugging
SDV($UA2MaxPermRecordCacheSize,  20);
SDV($UA2MaxPermQueryCacheSize,   100);

SDV($GuestUsergrp,               "GuestUsers");
SDV($LoggedInUsergrp,            "LoggedInUsers");
SDV($UA2UserPermDir,             "cookbook/userauth2/userperms"); 
SDV($UA2GroupPermDir,            "cookbook/userauth2/groupperms"); 
SDV($UA2ProfileDir,              "cookbook/userauth2/profiles");
SDV($UA2IpRangesDir,             "cookbook/userauth2/ipranges");  
SDV($UA2PermUpdateTimestampFile, "cookbook/userauth2/lastPermUpdateTimestamp"); 
  // contains the unix time the permissions were changed the last time

SDV($LoginPage,                  "Site.Login");
SDV($HomePage,                   "Home.Home");
SDV($UA2AfterSILoginRedirectTo,  ''); // default '' = previous content page
SDV($UA2AfterLogoutRedirectTo,   $LoginPage);
SDV($UA2AdminLoginFromIpsOnly,   ''); // default '' = no restriction

SDV($UA2AllowCookieLogin,        false);
SDV($UA2CookiePrefix,            ''); 
SDV($UA2CookieExpireTime,        60*60*24*30); // cookie default expiration in seconds
SDV($UA2LoginLockMsgFmt,         ''); // If set to a non-empty string, the module will block any login attempts
                                      // and display the string as message instead. For system maintainance

SDV($UA2AllowMultipleGranters,   false); // this option is honoured only on the UI level
SDV($UA2PermEditLockTimeout,     60*60*12); // time after which an edit lock on permission tables
                                            // is released forcefully when requested by someone else, 12 hours

SDV($UA2LoggedInUsersReplacements, 'replaceAuthIdInRecord'); // must be empty string or function (&$permrecord, $username)

// Note the difference between (pmwiki) actions and user auth specific admin actions 
// (which are all subsumed under pmwiki action 'admin').
SDVA($AdminActionsArr, array(
 // admin_action
  'report',
  'adduser',
  'addgroup',
  'edituser',
  'editgroup',
  'deluser',
  'delgroup',
  'setipperms'
));
// add some userauth specific pmwiki actions and levels
SDVA($HandleAuth, array(
 // action => level
  'admin'            => 'admin',
  'pwchange'         => 'pwchange',
  'pwset'            => 'pwset',
  'profile'          => 'profile',
  'diff'             => 'history', // this is a redefinition of the diff action to make it separately grantable
 // we have to take up the following (to pmwiki irrelevant) action-level pairs since otherwise, CheckUserPerms
 // will complain about unknown actions/levels
  'edituserperms'    => 'edituserperms',
  'createuserperms'  => 'createuserperms',
//  'editgugroup'      => 'editgugroup',
//  'editlugroup'      => 'editlugroup',
  'setipperms'       => 'setipperms'
));
$UA2AllLevels = array_values($HandleAuth); // make sure all action/level mappings are done before this point
SDVA($UA2AlwaysAllowedLevels, array(
  'logout'
));
SDVA($UA2LevelToAbbr, array(
 // level => abbreviation (used in perm table)
 // (must be bijection)
 // page related:
  'read'             => 'rd',
  'edit'             => 'ed',
  'upload'           => 'up',
  'history'          => 'hi',
 // non page related:
  'admin'            => 'ad', // may use admin tool (e.g. to change user rights) to the well-defined extent (see spec)
  'pwchange'         => 'pw', // user may change his own password
  'pwset'            => 'ps', // may set passwords for users below
  'profile'          => 'pr', // may view and alter his profile (excluding his password)
  'createuserperms'  => 'cu', // may create and delete users/groups
  'edituserperms'    => 'eu', // may edit users'/groups' permissions
  'setipperms'       => 'ip', // may change ip related permissions
//  'editgugroup'      => 'gu', // may edit guestusergrp permissions
//  'editlugroup'      => 'lu'  // may edit loggedinusergrp permissions
));
// which levels make only sense in connection with a page: (the other ones are usually user related)
SDVA($UA2PageRelatedLevelAbbr, array(
  'rd', 'ed', 'up', 'hi'
));

SDV($TempUnavailFmt,              "$[Server temporarily unavailable.]");
SDV($AuthRequiredFmt,             "$[Authentication required.]");
SDV($LackProperAbilitiesFmt,      "$[Insufficient privileges to perform action.]");
SDV($NoUsernameFmt,               "$[No username was provided.]");
SDV($InvalidUsernameFmt,          "$[Invalid username was provided.]");
SDV($WrongUserFmt,                $WrongPasswordFmt); // hide reason
//SDV($WrongUserFmt,                "$[Username does not exist.]");
SDV($WrongPasswordFmt,            "$[Wrong password supplied.]");
SDV($WrongIpRangeFmt,             "$[Wrong password supplied.]"); // hide reason
SDV($UnableToLoadPermRecordFmt,   $TempUnavailFmt);
SDV($UnableToLoadProfileFmt,      $TempUnavailFmt);
SDV($UnableToSetSessionCookieFmt, '$[Unable to set session cookie.]');
SDV($SessionInconsistencyFmt,     "$[Session client coupling inconsistent. Stopping for security reasons.]");
SDV($UA2SessionExpired1Fmt,       "$[Session has expired due to maximal life time or inactivity. Please log in again.]");
SDV($UA2SessionExpired2Fmt,       "$[Session has expired. Please log in again to complete the request.]");

$HandleActions['logout']       = 'HandleLogout';
$HandleActions['login']        = 'HandleLogin';
$HandleActions['pwchange']     = 'HandlePasswordChange'; // function defined in userauth2-pwchange.php

function getCurrClientIp() { return $_SERVER["REMOTE_ADDR"]; }
  // wrapped for testing purposes

// point the pmwiki permission checking mechanism to us:
$DefaultAuthFunction = $AuthFunction;
$AuthFunction = "UserAuth2";

// #013: increase PHP-builtin session timeouts
ini_set('session.cache_expire', max($UA2SessionMaxLifeTime/60, ini_get('session.cache_expire')));
ini_set('session.gc_maxlifetime', max($UA2SessionMaxLifeTime, ini_get('session.gc_maxlifetime')));
//ini_set('session.gc_probability', 1);
//ini_set('session.gc_divisor', 1); // no need to let PHP behave deterministic here - we do it ourself (see #014)

session_start(); // Note that it seems to be necessary to call session_start globally and before any occurence of
                 // (or reference to?) the _SESSION array. (Unfortunately this was undocumented.)
if(!isset($UserInstanceVars)) {
  require_once("userauth2/UserSessionVars.php");
  $UserInstanceVars = new UserSessionVars();
  if (!$UA2EnablePermCaching)
    $UserInstanceVars->clearPermCache(); // remove remains in old sessions if we have just switched off caching
}

//foreach ($_POST as $key => $value) {
//  debugMsg('', "POST start: $key => $value");
//}
//foreach ($_SESSION as $key => $value) {
//  debugMsg('', "SESSION start: $key => $value");
//}

function ua2OnSessionHijack($msg) { // needed some lines below
  global $UserInstanceVars, $SessionInconsistencyFmt;

  appendToUA2ErrorLog($msg . "\n");
  flushUA2ErrorLog();
  // Do not destroy session here, to leave the rightful owner undisturbed!
  // Rather, just do an id regenerate (without session file deletion!) such that we can safely proceed
  // to the login page (where the session with the then current id gets killed -- which is then not bad):
  session_regenerate_id(); // in PHP5 this would be session_regenerate_id(false);
  $UserInstanceVars->ClearInstanceVariables(); 
    // reset auth status for the new session (and no data carry-over!); this possibilizes also ip update later
  RedirectToLoginPage($SessionInconsistencyFmt);
}

// check whether client has ip address as memorized on session creation, to prevent malicious reuse/handover:
if (!isset($_SESSION['remote_addr'])) 
  $_SESSION['remote_addr'] =  getCurrClientIp();
else
  if ($UA2EnforceClientIpFixation && (getCurrClientIp() != $_SESSION['remote_addr'])) {
    ua2OnSessionHijack("Warning: Client ip mismatched memorized ip. Enforced session id renewal and auth reset.");
  }

// Check whether this session indeed belongs to our site: (important if we run this script on many sites of
// the same server, see www.php.net "session", comment gordon_e_rouse at ... dot com dot au, 29-Mar-2007 11:06.)
if (!isset($_SESSION['site_identifier']))
  $_SESSION['site_identifier'] = $UA2SiteIdentifier; 
else
  if ($UA2SiteIdentifier != $_SESSION['site_identifier']) {
    ua2OnSessionHijack("Warning: Site id mismatched memorized id. Enforced session id renewal and auth reset.");
  }

// do session expiration by hand since the PHP guys are just incapable to produce deterministic behaviour(*): #014
// ((*) Reason: www.php.net "session" shanemayer42 at ... dot com 20-Aug-2000 10:11)
$ua2Now = time();
if ((isset($_SESSION['lastrevivaltime']) &&  // inactivity time check
     ($ua2Now > $_SESSION['lastrevivaltime'] + $UA2SessionMaxInactivityTime)
    ) || 
    (isset($_SESSION['firststarttime']) &&   // max life time check
     ($ua2Now > $_SESSION['firststarttime'] + $UA2SessionMaxLifeTime)
    )
   ) 
{ 
  if ($UserInstanceVars->isAuthenticated()) {
    session_destroy(); session_start(); // (execute this only here, otherwise wouldn't know the login status anymore)
    $UserInstanceVars->setTargetUrl(getPageRequestUrl());
    $UserInstanceVars->setAuthMsg($UA2SessionExpired1Fmt);
    if (@$_POST['action'] && $UA2AllowPostCompletion) {
      $UserInstanceVars->setAuthMsg($UA2SessionExpired2Fmt);
      foreach($_POST as $k => $v) $_SESSION['POST_data'][$k] = $v;
    }
  } else {
    session_destroy(); session_start(); 
  }
}
// update times:
$_SESSION['lastrevivaltime'] = $ua2Now;
if (!isset($_SESSION['firststarttime'])) $_SESSION['firststarttime'] = $ua2Now;

// Revive possible posted data: since on session expiration we wander through a chain of 
// different page requests and redirections, use some conditions to make sure we revive the
// posted data only when we are sure we are at our aim.
if (isset($_SESSION['POST_data']) &&
    isContentPage($pagename) && ($action == @$_SESSION['POST_data']['action']) &&
    $UserInstanceVars->isAuthenticated()) {
  foreach($_SESSION['POST_data'] as $k => $v) $_POST[$k] = $v;
  unset($_SESSION['POST_data']);
}

// set some variables for pmwiki use:
$IsUserLoggedIn = $UserInstanceVars->isAuthenticated();
if ($UserInstanceVars->isAuthenticated()) {
  $AuthId = $UserInstanceVars->GetUsername();
  $Author = $UserInstanceVars->GetUsername();
    // It would be tempting to set $Author to the full name (meanwhile: description), but that 
    // may contain spaces etc. and could get bad beyond our userauth code.
}

//on initial loading of this code, set the calling stack to empty:
$UA2ChUsPmCallingStack = array();
//init some statistics vars:
$UA2StatUncachedRecordLoadsCount = 0; // both totals over complete page request
$UA2StatUncachedPermQueryCount = 0;

//Redirect exit handler: 
if (!isset($exitHandler)) $exitHandler = exit();
$UA2PreviousExitHandler = $exitHandler;
$exitHandler = "UA2myExitHandler";
  // Exit handler functionality will only show up sometimes unless all exits throughout pmwiki are
  // replaced by 'global $exitHandler; $exitHandler();'. It is only needed for statistics though.

function UA2myExitHandler() {
  global $THP_SESSION;
  global $UA2PreviousExitHandler;
  global $UA2StatUncachedRecordLoadsCount, $UA2StatUncachedPermQueryCount;

  appendToUA2ErrorLog("Current cache utilization: " .
                      count($_SESSION['permqueries']) . " perm queries, " .
                      count($_SESSION['userpermrecords']) . " user recs, ".
                      count($_SESSION['grouppermrecords']) . " group recs, " .
                      count($_SESSION['iprangerecords']) . " ip range recs.\n");
  appendToUA2ErrorLog("In total " . $UA2StatUncachedRecordLoadsCount . " uncached perm record loads, " .
                      $UA2StatUncachedPermQueryCount . " uncached perm queries.\n");

  flushUA2ErrorLog();

  // call original exit function:
  if ((strlen($UA2PreviousExitHandler) > 0) && function_exists($UA2PreviousExitHandler))
    $UA2PreviousExitHandler();
}

if (!function_exists(microtime_float)) {
  function microtime_float() {
    list($usec, $sec) = explode(" ", microtime());
    return ((float)$usec + (float)$sec);
  }
}

//================================================================
//======== Main entry point for pmwiki permission checking =======
//
// The process of accessing a page access is broken down into permission
// checking (functions that come later) and action taking like redirection
// etc. (done in the immediately following functions).

// First some channeling however for combatibility:
function UserAuth($pagename, $level, $authprompt=true) { 
  return UserAuth2($pagename, $level, $authprompt); 
}

function UserAuth2($pagename, $level, $authprompt=true) {
  // If called with $authprompt==true, the function is expected
  // to produce either the specified page to the client or 
  // print the login page. If called with $authprompt==false, the
  // caller just wants to know whether the current user has the
  // permission to access the specified page at the given level.
  // Still, the pmwiki engine demands the page being returned on 
  // success, since it needs the time stamps(?).
  if ($authprompt)
    return TryAccessingPage($pagename, $level);
  else {
    if (HasCurrentUserPerm($pagename, $level)) {
      $page = ReadPage($pagename);
      if (!$page) { return false; }
      return $page;
    }
  }
  return false;
}

function TryAccessingPage($pagename, $level) {
  // In this function the redirection logic is concentrated for the
  // cases of insufficient or wrong authentication. It returns
  // the desired page on success or appropriate auth dialogs otherwise.
  global $UserInstanceVars, $LackProperAbilitiesFmt, $AuthRequiredFmt;

  //appendToUA2ErrorLog("Someone trying to access page $pagename at level $level.\n");
  //appendToUA2ErrorLog("$pagename is a content page: " . answer(isContentPage($pagename)) . "\n");
  if (isContentPage($pagename)) {
    $UserInstanceVars->clearTargetUrl();
    // remember the (last step of the) trail of content pages the client is going: 
    $UserInstanceVars->setPrevContentPage($pagename);
  }
  if (HasCurrentUserPerm($pagename, $level)) {
    //appendToUA2ErrorLog("Access to $pagename at level $level granted.\n");
    // Return page:
    $page = ReadPage($pagename);
    if (!$page) { return false; }
    return $page;
  } else {
    //appendToUA2ErrorLog("Access to $pagename at level $level NOT granted.\n");
    if ($UserInstanceVars->isAuthenticated()) {
      PrintEmbeddedPageAndExit( $pagename, $LackProperAbilitiesFmt );
    } else {
      // Set redirection memory:
      $UserInstanceVars->setTargetUrl(getPageRequestUrl());

      // Send to login page...
      if ($UserInstanceVars->isAuthMsgSet()) // ... with auth message if already present, ...
        RedirectToLoginPage($UserInstanceVars->getAuthMsg());
 
      // ... or with standard message
      RedirectToLoginPage($AuthRequiredFmt); // won't return
    }
  }
  return false; // never reached
}

function getPageRequestUrl() {
  return ua2GetProtocol() . '://' . $_SERVER['HTTP_HOST'] . $_SERVER['REQUEST_URI'];
}

function ua2GetProtocol() {
  // Extra gimmick for choosing the right protocol for assembling the redirection url.
  // Is necessary when we start to handle SSL connections where not the whole site is
  // piped through the https protocol.

  // Use the protocol we are just using to access the wiki:
  if (@$_SERVER['HTTPS'] && ($_SERVER['HTTPS'] != 'off')) return 'https';
  return 'http';
}

//================================================================
//======== Permission checking ===================================
//
// The following functions are status-checkers, i.e. they take no action
// on the web-app state whatsoever. They however depend on the web-app state
// via the authentication status of the current user.

function HasCurrentUserPermForAction($page, $action) {
  global $HandleAuth;
  if (!isset($HandleAuth[$action])) {
    appendToUA2ErrorLog("Warning: Current user asking for permission for unknown action '$action'. Refused.\n");
    return false;
  }
  return HasCurrentUserPerm($page, $HandleAuth[$action]);
}

function HasCurrentUserPerm($page, $level) {
  // This is the (query-level) cached version of the function 
  // HasCurrentUserPermUncached() below.
  global $UserInstanceVars, $UA2EnablePermCaching;

  $UA2StartTime = microtime_float();

  if (!$UA2EnablePermCaching) {
    $res = HasCurrentUserPermUncached($page, $level);
    //appendToUA2ErrorLog("Executed in " . (microtime_float() - $UA2StartTime)*1000 . " msecs (uncached).\n");
    //flushUA2ErrorLog();
    return $res;
  }

  // look at the cache first:
  if ($UserInstanceVars->isPermQueryResCached($page, $level)) {
    $res = $UserInstanceVars->getPermQueryRes($page, $level); 
    //appendToUA2ErrorLog("Executed in " . (microtime_float() - $UA2StartTime)*1000 . " msecs (cached).\n");
    //flushUA2ErrorLog();
    return $res;
  }
  
  // or recalculate if not had been cached:
  $res = HasCurrentUserPermUncached($page, $level);

  // in any case cache it:
  $UserInstanceVars->pushPermQueryRes($page, $level, $res);

  //appendToUA2ErrorLog("Executed in " . (microtime_float() - $UA2StartTime)*1000 . " msecs (uncached).\n");
  //flushUA2ErrorLog();
  return $res;
}

function HasCurrentUserPermUncached($page, $level) {
  global $UserInstanceVars, $UA2StatUncachedPermQueryCount;

  $UA2StatUncachedPermQueryCount += 1;

  // First check by the current user's personal permission table:
  // (if here the username is undefined since not yet logged in, the
  //  parameter will be at least "as empty" such that #001 fails)
  if (CheckUserPerms($UserInstanceVars->GetUsername(), $page, $level)) 
    return true;

  // Otherwise try ip based (silent) granting:
  $ipRangePermRecords = $UserInstanceVars->getIpRangePermRecords();
  if ($ipRangePermRecords) 
    foreach($ipRangePermRecords as $ipRange => $permrecord) {
      if (ipMatches($ipRange, getCurrClientIp()) &&
          CheckUserPermsWithRecord($page, $level, $permrecord))
        return true;
    }
  
  return false;
}

//  The following functions perform web-app-state blind permission
//  checking: all what determines their result is fed through the parameters.

function CheckUserPermsForAction($user, $page, $action, $groupaction = false) {
  global $HandleAuth;
  if (!isset($HandleAuth[$action])) {
    appendToUA2ErrorLog("Warning: '$user' asking for permission for unknown action '$action'. Refused.\n");
    return false;
  }
  return CheckUserPerms($user, $page, $HandleAuth[$action], $groupaction);
}

function CheckUserPerms($user, $page, $level, $groupaction = false, $origuser = false) {
  global $UA2ChUsPmCallingStack;

  if (!$origuser) $origuser = $user; // #011

  // Make sure we dont cycle: (stop defensively if so) 
  // (important for complex group dependencies)
  if (isset($UA2ChUsPmCallingStack[$user]))
    return false;
 
  $UA2ChUsPmCallingStack[$user] = 1;
  $res = CheckUserPermsWorker($user, $page, $level, $groupaction, $origuser);
  unset($UA2ChUsPmCallingStack[$user]);
  if (empty($UA2ChUsPmCallingStack)) {
    if (false) // log PERM results?
    appendToUA2ErrorLog("Perm for user '$user', page '$page', level '$level': " .
                        ($res ? "passed.\n" : "failed.\n"));
    flushUA2ErrorLog(); // since every permission check passes this point exactly once
                        // (after having ascended from the recursion), do it here
  }
  return $res;
}

function CheckUserPermsWorker($user, $page, $level, $groupaction, $origuser) {
  // This function checks on permissions solely due to the
  // identification as given by the user name. If $user == '' or $user == false, 
  // then this counts as not logged in. If non-empty, expects a valid user string 
  // as input. Argument $page can be the empty string if $level is not page-related.
  global $LoginPage, $HomePage, $GuestUsergrp, $LoggedInUsergrp,
         $UserInstanceVars, $UA2AlwaysAllowedLevels, $UA2AllLevels,
         $UA2LoggedInUsersReplacements;

  if (!in_array($level, $UA2AllLevels)) {
    appendToUA2ErrorLog("Warning: Someone asking for permission for unknown level '$level'. Refused.\n");
    return false;
  }

  //appendToUA2ErrorLog("CheckUserPerms user $user page $page level $level...\n");

  // First the hardwired permissions:
  if ($level == 'attr') return false; // frequently occurring legacy level silently refused
  if (!$groupaction && ($user == 'admin')) return true;
  if (($level == 'read') && ($page == $LoginPage)) return true;
  //if (($level == 'read') && ($page == $HomePage)) return true; // see #010 in docu
  if (in_array($level, $UA2AlwaysAllowedLevels)) return true;
  if (!$groupaction) { // see #004 in docu
    if (CheckUserPerms($GuestUsergrp, $page, $level, true, $origuser)) return true; // #011 in docu
    if ((strlen($user) > 0) &&  // See #001
        CheckUserPerms($LoggedInUsergrp, $page, $level, true, $origuser)) return true;
  }

  if (strlen($user) == 0) return false;
  if (! doesUserExist_expected($user, $groupaction)) return false; 
    // here, for groups, return silently since someone could have deleted a group and the
    // perm tables haven't been updated by the admin yet; for users, chances are that 
    // we are checking for the parent of an orphaned user, so this is told in the error log


  // Then by saved record: (note that in any case the result that is obtained
  // via getPermHolderRecord is cached in the UserInstance, be it the guest user
  // logged in user record or the record of the user himself)
  $permrecord = $UserInstanceVars->getPermHolderRecord($user, $groupaction);
  if (($user == $LoggedInUsergrp) && function_exists($UA2LoggedInUsersReplacements))
    $UA2LoggedInUsersReplacements($permrecord, $origuser);
  if ($permrecord &&
      CheckUserPermsWithRecord($page, $level, $permrecord)) {
    return true;
  }

  // deny all otherwise:
  return false;
}

function CheckUserPermsWithRecord($page, $level, $permrecord) {
  // In this function the deeper internal logic of the delegation
  // scheme is concentrated. Two versions: 
  // (1) everyone can grant to everyone as much rights as he possesses 
  //   himself (a strategic world, google rank world), or 
  // (2) everyone in the line between admin and subject only can grant the 
  //   rights he possess himself (hierarchical world). 
  // Implemented is version (2).
 
  $parent = $permrecord['parent'];
  if (strlen($parent) == 0) return false; // empty parent would be a serious
                                          // inconsistency, so deny all

  $patronArr = getAllPatrons($parent); // is a proper user
  if ($patronArr === false) return false; // the parent or one of its patrons did not exist, so our
                                          // subject is an orphan => deny all.

  foreach($permrecord['perms'] as $granter => $permtable) {
    // First check if granter is patron of $subject: (BTW, make sure that all granters are proper users)
    // (take this out if you want version (1))
 
    if (($granter != 'admin') && !isInArrOfPatrons($granter, $patronArr)) continue;
      // This is equivalent to if (!isEqualOrPatronOf($granter, $parent)) continue;, only that
      // here we dont recalculate the patron array again and again.

    // Then check if granter has enough rights:
    if (! CheckUserPerms($granter, $page, $level)) continue;

    // Finally let's see how much he grants:
    if (CheckUserPermsWithTable($page, $level, $permtable, $granter))
      return true;
  }
  return false;
}

function CheckUserPermsWithTable($page, $level, $permtable, $granter) {
  // We expect here perm tables with each of the group membership specs
  // appearing before any elementary entry (see #002)

  // Honour group memberships on the fly, and transparently to the effects of
  // the elementary entries:
  $result = false; // default "deny all"
  foreach($permtable as $permentry) {
    if ($permentry{0} == '#') continue; // this was a comment
    if ($permentry{0} == '@') {         // this was a group membership spec
      $usergroup = substr($permentry, 1);
      if (!isPatronOf($granter, $usergroup, true)) continue; // (see docu #201)
      if (CheckUserPerms($usergroup, $page, $level, true)) $result = true; // see #205
    } 
    if ($permentry == '*') $result = true;
    if (entryApplicable($permentry, $page, $level)) {
      if ($permentry{0} == '-') 
        $result = false;
      else
        $result = true;
    }
  }

  return $result;
}

function entryApplicable($entry, $page, $level) {
  // Allowed parameters for $page: "group.page" or "group.*"
  // $entry must be of form "ad", "ed_group.page", "-ed_group.*", etc.
  //
  // Return values:
  // queried page:            entry:                     Result:
  // mygroup.mypage           mygroup.mypage             match
  // mygroup.mypage           *.mypage                   match
  // mygroup.mypage           mygroup.*                  match
  // mygroup.mypage           *.*                        match
  // mygroup.*                mygroup.*                  match
  // mygroup.*                *.*                        match
  // mygroup.*                mygroup.*d*                not match
  // mygroup.*                mygroup.anypage            not match
  //
  // When adding new features to the perm table syntax, make sure you update
  // isConsistentPermTable() also.
  global $UA2LevelToAbbr, $UA2PageRelatedLevelAbbr;

  if (strlen($entry) < 2) return false;
  if ($entry{0} == '-') $entry = substr($entry, 1);
  $levelspec = substr($entry, 0, 2); // first two characters
  $pagespec  = substr($entry, 3);    // everything after third character
  $levAbbr = $UA2LevelToAbbr[$level];
  if (!$levAbbr) 
    appendToUA2ErrorLog("Warning: Someone passed invalid permission level '$level' as argument. Refused.\n"); 
  if (!($pagespec && ($levelspec == 'xx') &&              // a joker level entry must be page related and must match
        in_array($levAbbr, $UA2PageRelatedLevelAbbr))) {  // a page related query;
    if ($levAbbr !== $levelspec) return false;            // otherwise compare levels as usual
  }
  if ($pagespec) { // if the entry is indeed page related (non-empty)
    if (! isValidQueriedPagename($page)) return false; // then be cautious
    return pagePatternMatches($pagespec, $page);
  }
  return true;
}

function getPagePart($page) {
  return substr($page, strpos($page, '.') + 1);
}

function getGroupPart($page) {
  return substr($page, 0, strpos($page, '.'));
}

function isGroupReference($page) {
  // returns true if $page is of form "group.*"
  return getPagePart($page) === '*';
} 

function pagePatternMatches($pattern, $pagename) {
  // $pattern must be of the form 'a??.d*e*' for example, $pagename must be
  // a valid queried page name. A '*' will match any number of any allowed characters,
  // the '?' will match exactly one allowed character.

  // At this point, all variables (only {$AuthId} at the moment) must be replaced already,
  if (strpos($pattern, '{') !== false) return false; // otherwise deny silently.

  if (strpos($pagename, '*') !== false) { // if is a group query then 
    if (isGroupReference($pattern)) { // if the patterns ends on '.*'
      $pagename = getGroupPart($pagename);  // then go for it, using only the group parts
      $pattern = getGroupPart($pattern);
    } else                                 // otherwise it cant be counted as a group related specifier
      return false;
  }
  // assemble reg exp out of page name pattern:
  $search = array('.', '?', '*');
  $replace= array('\.', '.', '(.*)');
  $regexp = str_replace($search, $replace, $pattern);
  // and fire:
  return preg_match('/^'.$regexp.'$/', $pagename); // case-sensitive
}

function replaceAuthIdInRecord(&$permrecord, $origuser) {
  // replaces all occurences of "{$AuthId}" in the perm tables of the given record by the
  // upper-cased version of the given user name.
  $newperms = array();
  foreach($permrecord['perms'] as $granter => $permtable) {
    $newperms[$granter] = str_replace('{$AuthId}', ucfirst($origuser), $permtable);
  }
  $permrecord['perms'] = $newperms;
}

//====================================================
//======== Validity checks ===========================

function isValidUserString($user, $groupaction = false) {
  // user string may contain only letters, digits and underscore, and
  // must begin with a letter; ensure that it is a valid file name on your
  // system!
  return preg_match('/^[a-zA-Z][a-zA-Z0-9\_]*$/', $user);
    // (for both users and groups the same requirement)
}

function isValidChosenUserString($user, $groupaction = false) {
  // This function does not decide regarding whether the user already exists!
  global $GuestUsergrp, $LoggedInUsergrp;

  // first, the string must be a valid choice:
  if (!isValidUserString($user, $groupaction)) return false;

  // since the edit lock files are constructed by prepending a dot to the
  // user name, some choices are rather forbidden as user/group names:
  if (strcasecmp($user, 'htaccess') == 0) return false;
  if (strcasecmp($user, 'htpasswd') == 0) return false;
  if (strcasecmp($user, 'htgroup') == 0) return false;

  // some are just inappropriate (for example to similar to special names):
  if ($groupaction && ($user === $GuestUsergrp)) return true; // exact matches to guest or logged in user are ok
  if ($groupaction && ($user === $LoggedInUsergrp)) return true; 
  if (strcasecmp($user, $GuestUsergrp) == 0) return false;    // mimicks of these are forbidden
  if (strcasecmp($user, $LoggedInUsergrp) == 0) return false;
  if (strcasecmp($user, 'admin') == 0) return false; // admin user and its mimicks are tabu

  return true;
}

function isValidIpRange($ipRange) {
  // This function is needed for deciding whether
  // to save a group into the group perms or ip ranges folder.
  // Returns true if argument is of the form 
  // 192.
  // 192.168.
  // 192.168.1.
  // 192.168.1.10
  // false otherwise. 
  if (!preg_match('/^[0-9\.]+$/', $ipRange)) return false;
  $buf = explode(".", $ipRange);
  if (count($buf) < 2) return false;
  if (count($buf) > 4) return false;
  $flag = false; // the construction with the flag makes any empty
                 // byte field which is not the last to cause a neg answer
  foreach($buf as $byteVal) {
    if ($flag) return false;
    if (strlen($byteVal) == 0)
      $flag = true;
    else
      if (($byteVal < 0) || ($byteVal > 255)) return false;
  }
  return true;
}

function isIpRange($alleged_iprange) {
  // This is the meager version of the function above.
  // It expects a valid perm holder string, i.e. valid ip range or
  // valid user/group string.
  if (($alleged_iprange{0} >= '0') && ($alleged_iprange{0} <= '9'))
    return true;
  return false;
}

function isValidPermHolderString($user, $groupaction = false) {
  return isValidUserString($user, $groupaction) ||
         ($groupaction && isValidIpRange($user));
}

if (!function_exists(lcfirst)) {
function lcfirst($str) { // like ucfirst()
  if (strlen($str) == 0) return '';
  return strtolower($str{0}) . substr($str, 1);
}
}

function isValidChosenPermHolderString($user, $groupaction = false) {
  if (doesPermHolderExist(ucfirst($user), $groupaction, false) ||
      doesPermHolderExist(lcfirst($user), $groupaction, false))
    return false; // see #012 in docu

  return isValidChosenUserString($user, $groupaction) ||
         ($groupaction && isValidIpRange($user));
}

function isValidFullPagename($pagename) {
  // Just for reference, to explicitize that valid pages contain only
  // letters and digits, and group and page part must be non-empty.
  return preg_match('/^[a-z0-9]+\.[a-z0-9]+$/i', $pagename);
}

function isValidQueriedPagename($pagename) {
  // Returns true if the argument is valid when provided to CheckUserPerms etc.
  // as $page argument. Currently only full qualified pages (MyGroup.MyPage) or
  // group queries (MyGroup.*) are permitted.
  return preg_match('/^[a-z0-9]+\.\*|([a-z0-9]+)$/i', $pagename);
}

function isValidPagenamePattern($pattern) {
  // For validity check of page specifiers in page related perm entries.
  // Returns true if it contains exactly one dot, and group and page part are non-empty,
  // consist of letters, digits, '*' and '?' only and each dont start with a digit.
  // Examples are: 'a.b', '*a*.h12???', '*.*', 'dfsf.*', '*.df' and so on.
  $pattern = str_replace('{$AuthId}', 'Max', $pattern);
  return preg_match('/^[a-z0-9\?\*]+\.[a-z0-9\?\*]+$/i', $pattern);
  //return preg_match('/^[a-z\?\*][a-z0-9\?\*]*\.[a-z\?\*][a-z0-9\?\*]*$/i', $pattern); // old
}

function isContentPage($pagename) {
  global $LoginPage;

  return isValidFullPagename($pagename) &&
         ($pagename != $LoginPage);
}

//=====================================================
//======== File handling ==============================

function getUserPermFilename($user) {
  global $UA2UserPermDir;

  return "$UA2UserPermDir/$user";
}

function getGroupPermFilename($group) {
  global $UA2GroupPermDir;

  return "$UA2GroupPermDir/$group";
}

function getIpRangePermFilename($iprange) {
  global $UA2IpRangesDir;

  return "$UA2IpRangesDir/$iprange";
}

function getPermHolderFilename($user, $groupaction = false) {
  // Generalized version of the above three. Expects $user to be
  // either a valid ip range or valid user/group string.
  if ($groupaction) {
    if (isIpRange($user)) 
      return getIpRangePermFilename($user);
    else
      return getGroupPermFilename($user);
  }
  return getUserPermFilename($user);
}

function getEditLockFilename($user, $groupaction = false) {
  // Generalized version.
  // Practically, these hidden files are automatically ignored when reading users/groups.
  if ($groupaction) {
    if (isIpRange($user))
      return getIpRangePermFilename('.' . $user);
    else
      return getGroupPermFilename('.' . $user);
  }
  return getUserPermFilename('.' . $user);
}

function doesUserExist($user, $groupaction = false, $use_cache = true) {
  global $UserInstanceVars;

  if (!$groupaction && ($user == 'admin')) return true;

  if ($use_cache && $UserInstanceVars->isUserPermRecordCached($user, $groupaction))
    return true;

  if ($groupaction) return doesGroupExist($user);

  return file_exists(getUserPermFilename($user));
}

function doesUserExist_expected($user, $groupaction = false) {
  // This is the same version as before, but with error reporting
  // included for cases where we expect the user to exist, but it
  // doesn't. For groups we go silent, since it is easier to forget
  // to delete all group references and it is less critical. (On the
  // other hand a user without parent is a orphan in the system,
  // can be manipulated by the admin only.)

  if (doesUserExist($user, $groupaction)) return true;
  if (!$groupaction)
    appendToUA2ErrorLog("Warning: Expected user '$user' did not exist. (could be ghost parent of an orphan)\n");
  return false;
}

function doesGroupExist($group) {
  return file_exists(getGroupPermFilename($group));
}

function doesIpRangeExist($iprange) {
  return file_exists(getIpRangePermFilename($iprange));
}

function doesGenGroupExist($group) {
  return doesGroupExist($group) || doesIpRangeExist($group);
}

function doesPermHolderExist($user, $groupaction = false, $use_cache = true) {
  // This is a permrecord-cached version valid for all three types
  // of "users" (user, group, iprange).
  // Expects $user to be a valid perm holder string, i.e. either
  // a valid ip range, or a valid user or group name. Distinction
  // between ip range and the others is made via the first character
  // in $user, distinction between user and group name is made via 
  // $groupaction.
  global $UserInstanceVars;

  if ($groupaction && isIpRange($user)) {
    if ($use_cache && $UserInstanceVars->isIpRangeRecordCached($user, $groupaction))
      return true;
    return doesIpRangeExist($user);
  }

  return doesUserExist($user, $groupaction, $use_cache);
}

function loadUserPermRecord($user, $groupaction = false) {
  if (strlen($user) == 0) return false;

  //appendToUA2ErrorLog("Loading perm record for $user.\n");

  if ($groupaction)
    return loadPermRecord_(getGroupPermFilename($user));
  else
    return loadPermRecord_(getUserPermFilename($user));
 
  return false;
}

function loadIpRangePermRecord($iprange) {
  if (strlen($iprange) == 0) return false;

  //appendToUA2ErrorLog("Loading perm record for $iprange.\n");

  return loadPermRecord_(getIpRangePermFilename($iprange));
}

function loadPermHolderRecord($holder, $groupaction) {
  // Expects a valid holder string.
  if ($groupaction && isIpRange($holder))
    return loadIpRangePermRecord($holder);
  else
    return loadUserPermRecord($holder, $groupaction);
  return false;
}

function loadPermRecord_($file) {
  global $UA2StatUncachedRecordLoadsCount;
  $UA2StatUncachedRecordLoadsCount += 1;
  $fp = @fopen($file, 'r');
  if ($fp && flock($fp, LOCK_SH)) { // if user exists, load record
    $sz = filesize($file);
    $permrecord = unserialize(fgets($fp, $sz));
    flock($fp, LOCK_UN);
    fclose($fp);
    return $permrecord;
  }
  else
    appendToUA2ErrorLog("Could not open permission record at $file for reading.\n");
  return false;
}

function saveUserPermRecord($user, $record, $groupaction = false) {
  if ($groupaction)
    return savePermRecord_(getGroupPermFilename($user), $record);
  else
    return savePermRecord_(getUserPermFilename($user), $record);
}

function saveGroupPermRecord($group, $record) {
  return savePermRecord_(getGroupPermFilename($group), $record);
}

function saveIpRangePermRecord($iprange, $record) {
  return savePermRecord_(getIpRangePermFilename($iprange), $record);
}

function savePermHolderRecord($holder, $record, $groupaction) {
  // This is the most generalized form, with holder being
  // either a normal user (groupaction==false) or a normal group
  // or ip range (identified from the holder string; both with 
  // groupaction==true).
  if (strlen($holder) == 0) { 
    appendToUA2ErrorLog("Warning: Tried to save to empty user name. Prevented.\n");
    return false;
  }
  if ($groupaction && isIpRange($holder))
    return saveIpRangePermRecord($holder, $record);
  else
    return saveUserPermRecord($holder, $record, $groupaction);
}

function savePermRecord_($file, $record) {
  $fp = @fopen($file, "w");
  if ($fp && flock($fp, LOCK_EX)) {
    fputs($fp, serialize($record) . "\n");
    flock($fp, LOCK_UN);
    fclose($fp);
    return true;
  }
  return false;
}

function delUser($user, $groupaction = false) {
  global $UA2ProfileDir;

  if ($groupaction) 
    @unlink(getGroupPermFilename($user));
  else {
    @unlink(getUserPermFilename($user));
    @unlink("$UA2ProfileDir/$user");
  }
}

function delPermHolder($holder, $groupaction) {
  if ($groupaction && isIpRange($holder))
    @unlink(getIpRangePermFilename($holder));
  else
    delUser($holder, $groupaction);
}

function loadIpRangePermRecords() {
  global $UA2IpRangesDir;
  
  $res = array();
  $dirp = @opendir($UA2IpRangesDir);
  if (!$dirp) return $res;
  while (($file=readdir($dirp)) !== false) {
    if ($file[0]=='.') continue;
    $res[$file] = loadIpRangePermRecord($file);
  }
  closedir($dirp);
  return $res;
}

function getAllUserPermRecords() {
  global $UA2UserPermDir;

  $res = array();
  $dirp = @opendir($UA2UserPermDir);
  if (!$dirp) return $res;
  while (($user=readdir($dirp)) !== false) {
    if ($user[0]=='.') continue;
    $res[$user] = loadPermRecord_(getUserPermFile($user));
  }
  closedir($dirp);
  return $res;
}

function getAllGroupPermRecords() {
  global $UA2GroupPermDir;

  $res = array();
  $dirp = @opendir($UA2GroupPermDir);
  if (!$dirp) return $res;
  while (($group=readdir($dirp)) !== false) {
    if ($group[0]=='.') continue;
    $res[$group] = loadPermRecord_(getGroupPermFile($group));
  }
  closedir($dirp);
  return array_merge($res, loadIpRangePermRecords());
}

function loadUserProfile($user) {
  global $UA2ProfileDir;

  $filename = "$UA2ProfileDir/$user";

  $fp = @fopen($filename, 'r');
  if ($fp) { // if user exists, load profile
    $sz = filesize($filename);
    $profile = unserialize(fgets($fp, $sz));
    fclose($fp);
    return $profile;
  }
  else
    appendToUA2ErrorLog("Could not open profile at $filename for reading.\n");
  return false;
}

function saveUserProfile($user, $profile) {
  global $UA2ProfileDir;
  
  $filename = "$UA2ProfileDir/$user";

  $fp = @fopen($filename, 'w');
  if ($fp) { // if successful
    fputs($fp, serialize($profile) . "\n");
    fclose($fp);
    return true;
  }
  else 
    appendToUA2ErrorLog("Could not open profile at $filename for writing.\n");
  return false;
}

function storeCookieKey($user, $cookiekey, $keycreatetime) {
  $profile = loadUserProfile($user);
  if (!$profile) return false; // fail silently if profile could not be loaded
  $profile['cookiekey'] = $cookiekey;
  $profile['cookiekeycreatetime'] = $keycreatetime;
  return saveUserProfile($user, $profile);
}

function getCookieKey($user, $max_age) { 
  // Returns false if the cookie key could not be loaded or has expired, otherwise
  // the cookie key.
  $profile = loadUserProfile($user);
  if (!$profile) return false; 
  if ($profile['cookiekeycreatetime'] + $max_age < time()) return false;
  return $profile['cookiekey'];
}

//==============================================================
//======== Ip address related helpers ==========================

function ipMatches($ipRange, $client_ip) {
  // Checks whether the given client ip matches the ip range given in $ipRange.
  // For $ipRange strings as "192.168.1.", "134.121.34.2", or "10.1." are allowed
  // (take care to use the final dot for actual ranges!).
  // $ipRange may be a comma-separated list of those strings in the future.
  if (count(explode(".", $ipRange)) == 4) // if iprange is a full ip address,
      // set an end marker to avoid 192.168.1.112 matching iprange = 192.168.1.1 (e.g.)
    return (strpos("x" . $client_ip . ",", $ipRange . ",", 1) == 1);
  return (strpos("x" . $client_ip, $ipRange, 1) == 1);
  // Note the begin marker "x" and testing on == 1 is also for the reason to avoid
  // matching the range "92." on an ip address "192.168.1.1".
}

function ipMatchesIpRangeArr($ipRangeArr, $client_ip) {
  // Returns true is the specified ip matches at least one of the
  // given ip ranges in $ipRangeArr.
  foreach($ipRangeArr as $ipRange)
    if (ipMatches($ipRange, $client_ip))
      return true;
  return false;
}

function currClientIpMatchesRange($ipRange) {
  return (boolean)ipMatches($ipRange, getCurrClientIp());
}

//==============================================================
//======== Output related functions ============================

function GetUserLoginForm($no_title = true) {
  global $pagename, 
         $AuthId,
         $WikiTitle,
         $UserInstanceVars,
         $UA2AllowCookieLogin, 
         $UA2CookieExpireTime;

  $login_form = '';

  if (!$UserInstanceVars->isAuthenticated()) {

    if (!$no_title) {
      $login_form .= "<h1>$[Login to] $WikiTitle</h1>";
    }

    if (strlen($UserInstanceVars->getAuthMsg()) > 0) {
      $login_form .= $UserInstanceVars->getAuthMsg() . "<br>&nbsp;<p>\n";
      $UserInstanceVars->setAuthMsg('');
    }

    $login_form .=
      "<form name='authform' action='" . $_SERVER['REQUEST_URI'] . "?action=login' method='post'>
       <table class='userauthtable' style='padding:0px; margin:0px'>

       <tr>
           <td>$[Username]:</td>
           <td><input tabindex='1' name='username' class='userauthinput' value='' /></td>
       </tr>

       <tr>
           <td>$[Password]:</td>
           <td><input tabindex='2' name='password' class='userauthinput' type='password' value='' /></td>
       </tr>";

    // setting table style paddings and margins to zero has the effect that the table will
    // actually be left-justified !!! (otherwise this is forgotten by the browser)

    if($UA2AllowCookieLogin) {
      $login_form .= "
       <tr height=40px>
           <td>$[Keep me logged in<br> on this browser]:</td>
           <td><input tabindex='3' name='persistent' class='userauthinput' type='checkbox' value='1' /> &nbsp; ($[for] " .
            $UA2CookieExpireTime/86400 . " $[days])</td>
       </tr>";
    }

    $login_form .= "
       <tr height=35px>
           <td align=left><input tabindex='3' class='userauthbutton' type='submit' value='$[Login]' /></td>
           <td>&nbsp;</td>
       </tr>
       </table>
       </form>
       <script language='javascript' type='text/javascript'><!--
          document.authform.username.focus() //--></script>";

  } else {
    $login_form =
      "$[Logged in as]: " . $AuthId . "<br><br>
       <a href=\"" . $_SERVER['REQUEST_URI'] . "?action=logout\">$[Logout]</a>";

  }

  return FmtPageName($login_form, $pagename);
}

function PrintEmbeddedPageAndExit($pagename, $message) {
  global $PageStartFmt, $PageEndFmt;

  $page = ReadPage($pagename);
  PCache($pagename, $page);

  $AuthEmbeddedFmt = array( &$PageStartFmt,
                            $message,
                            &$PageEndFmt);

  PrintFmt($pagename, $AuthEmbeddedFmt);
  global $exitHandler; $exitHandler();
}

//=============================================================
//======= Authentication handling =============================

function doPasswordsMatch($provided_passwd, $stored_passwd) {
  // checks whether the password given as input by the user ($provided_passwd)
  // matches the stored one.
  if (($provided_passwd === '') && ($stored_passwd === '')) return true; // empty password
  return crypt($provided_passwd, $stored_passwd) == $stored_passwd;
}

function checkPasswordForUser($alleged_user, $alleged_passwd) {
  // $alleged_user must be a valid user name string.
  // If the stored password is accessible, this function returns 'success' if
  // the passwords matched. If the stored password is not accessible
  // this function returns a corresponding error message.
  global $WrongPasswordFmt,
         $UnableToLoadProfileFmt;

  // Method "profile": if (AuthMethod == profile) { ...
  $alleged_profile = loadUserProfile($alleged_user);
  if (! $alleged_profile) {
    appendToUA2ErrorLog("Could not check password since unable to load profile for $alleged_user.\n");
    return $UnableToLoadProfileFmt;
  }

  // Add other authentication methods here: if (AuthMethod == whatever) { ...

  if (!doPasswordsMatch($alleged_passwd, $alleged_profile['password']))
    return $WrongPasswordFmt;

  return 'success';
}

function setPasswordForUser($user, $new_password) {
  global $ProblemsSavingSettingsFmt;

  // if method==profile
  $profile = loadUserProfile($user);
  if ($profile) { // profile will exist since we just checked the old password against it
    if (strlen($new_password) == 0)
      $profile['password'] = '';
    else 
      $profile['password'] = crypt($new_password);
    if (saveUserProfile($user, $profile))
      return 'success';
  }
  return $ProblemsSavingSettingsFmt;
}

function my_session_regenerate_id($carryOverData = false) {
  // Adapted from post on www.php.net session_regenerate_id() by Nicholas 03/06/2005.
  // Deletes old session file, thus making the old session id completely invalid.
  // Pass $carryOverData == true if you want to retain the data in the session. 
  session_start();
  $old_sessid = session_id();
  session_regenerate_id();
  $new_sessid = session_id();
  session_id($old_sessid);
  session_destroy();

  //If you don't copy the $_SESSION array, you won't be able to use the data associated with the old session id.
  $old_session = $_SESSION;
  session_id($new_sessid);
  session_start();
  if ($carryOverData)
    $_SESSION = $old_session;

  $ua2Now = time();
  $_SESSION['firststarttime'] = $ua2Now;
  $_SESSION['lastrevivaltime'] = $ua2Now;

  return true;
}

function HandleLogin($pagename) {
  global $THP_SESSION;
  global $UserInstanceVars, 
         $UA2LoginLockMsgFmt,
         $UA2AllowCookieLogin,
         $UA2AdminLoginFromIpsOnly,
         $WrongIpRangeFmt,
         $WrongUserFmt,
         $UnableToLoadPermRecordFmt,
         $UnableToSetSessionCookieFmt;

  // all these redirects implicitly exit the php code
 
  if (strlen($UA2LoginLockMsgFmt) > 0)
    RedirectToLoginPage($UA2LoginLockMsgFmt);

  if (!isset($_REQUEST['username'])) 
    RedirectToLoginPage($NoUsernameFmt);

  $alleged_user    = $_REQUEST['username'];
  $provided_passwd = $_REQUEST['password'];
  //$THP_SESSION['alleged_username'] = $alleged_user; // might be useful to keep

  // For better protection against brute force attacks, the negative branches of the
  // following if statements might profit from including some delay lock, for example
  // by having a sleep() here and restricting the number of clients involved in a 
  // login action at a time.

  if (! isValidUserString($alleged_user))
    RedirectToLoginPage($InvalidUsernameFmt);

  appendToUA2ErrorLog("Handle login for alleged user '$alleged_user' from " . getCurrClientIp() . " ...\n");

  if (! doesUserExist($alleged_user))
    RedirectToLoginPage($WrongUserFmt);

  // user does indeed exists, so proceed to password checking etc.:

  // first check password 
  $check_res = checkPasswordForUser($alleged_user, $provided_passwd);
  if ($check_res != 'success') {
    RedirectToLoginPage($check_res);
  }

  // second check ip range 
  if ($alleged_user != 'admin') { // using the perm record if not admin
    $alleged_permrecord = $UserInstanceVars->getPermHolderRecord($alleged_user, false, false, false);
    if (! $alleged_permrecord)
      RedirectToLoginPage($UnableToLoadPermRecordFmt);

    if (!empty($alleged_permrecord['loginFromIpsOnly'])) // check only if set and non-empty, see #100
      if (! ipMatchesIpRangeArr($alleged_permrecord['loginFromIpsOnly'], getCurrClientIp()))
        RedirectToLoginPage($WrongIpRangeFmt);
  } else { // or using the special variable for admin
    if (strlen($UA2AdminLoginFromIpsOnly) > 0)
      if (! ipMatchesIpRangeArr($UA2AdminLoginFromIpsOnly, getCurrClientIp()))
        RedirectToLoginPage($WrongIpRangeFmt);
  }
  
  // check was successful, so try session regeneration:
  //old: if (thp_session_regenerate($THP_SESSION, true)) { // try to regenerate as authenticated
  if (my_session_regenerate_id(true)) { // generate new session id and destroy old one, but carry over data
    // if ok, then finish off
    $UserInstanceVars->SetUsername($alleged_user); // = "is authenticated"
    appendToUA2ErrorLog("User '$alleged_user' has logged in.\n");
    //appendToUA2ErrorLog("Session was regenerated.\n");
    $UserInstanceVars->clearPermQueryCache(); // make really really sure we work with new permissions

    if($UA2AllowCookieLogin && @$_REQUEST['persistent']) {
      $UserInstanceVars->setAuthCookie();
      appendToUA2ErrorLog("Authentication cookie has been set for user $alleged_user.\n");
    } else
      $UserInstanceVars->clearAuthCookie();

    RedirectOnSuccessfulLogin();
  } else {
    RedirectToLoginPage($UnableToSetSessionCookieFmt);
  }
}

function RedirectOnSuccessfulLogin() {
  // This function is just to wrap the redirection operations
  global $UserInstanceVars, $UA2AfterSILoginRedirectTo, $HomePage;

  // if the user had some target in mind, then redirect to that now ("finally"):
  if (($target_url_loc = $UserInstanceVars->getTargetUrl())) { // this is an assignment
    $UserInstanceVars->clearTargetUrl();
    RedirectToURL($target_url_loc); // exits with calling the pmwiki exit handler
  }

  // otherwise redirect to the by-option specified page:
  if (strlen($UA2AfterSILoginRedirectTo) > 0) { 
    // do some possibly fancy redirect (depending on the user name)
    Redirect(str_replace('{$AuthId}', $UserInstanceVars->GetUsername(), 
                         $UA2AfterSILoginRedirectTo));
  } 

  if (($prev_contentpage = $UserInstanceVars->getPrevContentPage())) { // again
    Redirect($prev_contentpage); // exits with calling the pmwiki exit handler
  }

  // otherwise try the home page
  if (HasCurrentUserPerm($HomePage, 'read'))
    Redirect($HomePage); // exits with calling the pmwiki exit handler
  else 
    PrintEmbeddedPageAndExit('Please make sure at least the home page is readable for a logged ' . 
                             'in user, or set $UA2AfterSILoginRedirectTo appropriately.'); // needs no i18n
}

function HandleLogout($pagename) {
  global $THP_SESSION;
  global $UserInstanceVars, 
         $UA2AfterLogoutRedirectTo;

  appendToUA2ErrorLog("User '" . $UserInstanceVars->GetUsername() . "' has logged out.\n");

  // clear authentication cookie:
  $UserInstanceVars->clearAuthCookie();

  // clear UserAuth instance variables:
  $UserInstanceVars->ClearInstanceVariables();
  //my_session_regenerate_id(false); 
  session_destroy();
  //thp_session_destroy($THP_SESSION);

  // redirect to by-option specified page if available:
  if (strlen($UA2AfterLogoutRedirectTo) > 0) {
    Redirect($UA2AfterLogoutRedirectTo); // exits with calling the pmwiki exit handler
  }

  // otherwise go to login page:
  Redirect($LoginPage); // exits with calling the pmwiki exit handler
}

//=============================================================
//======= Redirection helpers =================================

function RedirectToLoginPage($msg) {
  global $THP_SESSION;  
  global $LoginPage, $UserInstanceVars; 

  $UserInstanceVars->setAuthMsg($msg);
  Redirect($LoginPage); // exits with calling the pmwiki exit handler
}

function RedirectToURL($url) {
  global $RedirectDelay;
  clearstatcache();

  header("Location: $url");
  header("Content-type: text/html");
  print("<html><head>
    <meta http-equiv='Refresh' Content='$RedirectDelay; URL=$url'>
    <title>Redirect</title></head><body></body></html>");

  global $exitHandler; $exitHandler(); // like exit;
}

function appendToUA2ErrorLog($msg) {
  global $UA2ErrorLogLastMsg,
         $UA2ErrorLog;

  if ($msg != $UA2ErrorLogLastMsg) { // includes ('' != false) yielding false
    $UA2ErrorLogLastMsg = $msg;
    $UA2ErrorLog .= $msg;
  }
}

function flushUA2ErrorLog() {
  global $UA2ErrorLog;

  if ((strlen($UA2ErrorLog) > 0) && function_exists(debugMsg)) 
    debugMsg('USAU', $UA2ErrorLog);

  $UA2ErrorLog = '';
  // UA2ErrorLogLastMsg can stay set
}

//==============================================================
//======== Structural functions on the perm record and related 

// ------------ initialization and cleaning ------------------

function newEmptyProfile() {
  $prof['password'] = ''; // empty password as default
  return $prof;
}

function newEmptyPermRecord($parent, $groupaction = false) {
  $res['parent'] = $parent;
  if (!makeConsistent($res, $groupaction)) return false;
  return $res;
}

function getConsistent($permrecord, $groupaction = false) {
  if (!makeConsistent($permrecord, $groupaction)) return false;
  return $permrecord;
}

function makeConsistent(&$permrecord, $groupaction = false) {

  $parent = $permrecord['parent'];
  if ((strlen($parent) == 0) || !doesUserExist($parent)) return false; 

  // set loginFromIpsOnly field to a definite value if is not set,
  // but only if it is a proper user; for groups delete it:
  if ($groupaction)
    unset($permrecord['loginFromIpsOnly']);
  else
    if (!isset($permrecord['loginFromIpsOnly']))
      $permrecord['loginFromIpsOnly'] = array();

  // same with the perms array: have at least an empty perm table for the parent
  if (!isset($permrecord['perms']))
    $permrecord['perms'][$parent] = array();

  // remove all non-authorized permission granters (and check perm tables on the fly):
  if (!empty($permrecord['perms'])) {
    $patronArr = getAllPatrons($parent);
    if ($patronArr === false) return false;
    foreach($permrecord['perms'] as $granter => $permtable) { // traversing a changing array works
      if (($granter != admin) &&
          (!doesUserExist($granter) ||
           !isInArrOfPatrons($granter, $patronArr)
          )
         ) {
        appendToUA2ErrorLog("Warning: '$granter' granting though not patron or non-existent. I keep deleting him.\n");
        unset($permrecord['perms'][$granter]); 
      }
      if (!isConsistentPermTable($permtable, $granter)) return false;
    }
  }

  return $permrecord;
}

// ------------- patron related -------------------------------

function isEqualOrPatronOf($alleged_patron, $dependent, $groupaction = false) {
  // Returns whether $alleged_patron is equal to or a patron of $dependent.
  // $alleged_patron must be proper user, $dependent may be user or group
/*
  if (!$groupaction)
    if ($dependent == $alleged_patron) 
      return true;
  return isPatronOf($alleged_patron, $dependent, $groupaction);
*/
  if ($alleged_patron == 'admin') return true;
  $patronArr = getAllPatrons($dependent, $groupaction);
  return isInArrOfPatrons($alleged_patron, $patronArr);
} 

function isPatronOf($alleged_patron, $dependent, $groupaction = false) {
  // Returns whether $alleged_patron is a (strict) patron of $dependent.
  // $alleged_patron must be proper user, $dependent may be user or group
/*
  if ($dependent == 'admin') return false;
  if ($alleged_patron == 'admin') return true;
  $patronArr = getAllPatrons($dependent, $groupaction);
  return ($groupaction || ($dependent != $alleged_patron)) && 
         isInArrOfPatrons($alleged_patron, $patronArr);
*/
  if (!$groupaction && ($dependent == $alleged_patron)) return false;
  return isEqualOrPatronOf($alleged_patron, $dependent, $groupaction);
}

function getAllPatrons($user, $groupaction = false) {
  // Returns an associative array of $user and all its patrons, excluding 'admin'.
  // Entries are of form $patron => $thePatronsParent, so (provided consistent
  // perm records) the first included entry is xxx => admin and the last included
  // is $user => yyy.
  global $UserInstanceVars; // we make use of almost arbitrary caching here
                            // (data need not necessarily have to do with the
                            // user represented by the user instance)
  // fill up result array from bottom-up:
  // (Due to some practical consideration we do not include the root 
  // parent 'admin' in the array.)
  if (!$groupaction && ($user == 'admin')) return array();
  $permrecord = $UserInstanceVars->getPermHolderRecord($user, $groupaction);
  if ($permrecord === false) return false;
  $parent = $permrecord['parent'];
  $res = getAllPatrons($parent); // parent is definitely a user
  if ($res === false) return false;
  $res[$user] = $parent;
  return $res;
}

function getAllPatronsStrictlyBelow($patron, $user, $groupaction = false) {
  // Returns the array of patrons of $user where all entries higher than $patron
  // are removed.
  $buf = getAllPatrons($user, $groupaction);
  if ($buf === false) return false;
  $res = array();
  $pointer = $user;
  while (($pointer != false) && isset($buf[$pointer])) {
    array_unshift($res, $buf[$pointer]); // unshift the parent of the currently considered patron:w
    $pointer = $buf[$pointer];
  }
  // now we have got an array of parents, in descending order starting with admin, so remove
  // the ones that are above $patron now:
  while (array_shift($res) !== $patron) ;
  return $res;
}

function getAllPatronsEqualOrBelow($patron, $user, $groupaction = false) {
  $res = getAllPatronsStrictlyBelow($patron, $user, $groupaction);
  if ($res === false) return false;
  array_unshift($res, $patron);
  return $res;
}

function isInArrOfPatrons($patron, $patronArr) {
  return isset($patronArr[$patron]);
}

// ------------- lists ----------------------------------------

function getAllUsers($groupaction = false, $use_cache = true, 
                     $cache_it = true, $assure_consistency = true) {
  // Returns an array of all users (or groups and ip ranges) in the
  // form $user => $theUsersParent.
  global $UA2GroupPermDir, $UA2UserPermDir, $UserInstanceVars;

  $res = array();
  if ($groupaction) $dirp = @opendir($UA2GroupPermDir);
  else              $dirp = @opendir($UA2UserPermDir);
  if (!$dirp) return false;
  while (($user=readdir($dirp)) !== false) {
    if ($user[0]=='.') continue;
    $permrecord = $UserInstanceVars->getPermHolderRecord(
                    $user, $groupaction, $use_cache, $cache_it, $assure_consistency);
    $res[$user] = $permrecord['parent'];
  }
  closedir($dirp);
  if ($groupaction) {
    $buf = $UserInstanceVars->getIpRangePermRecords();
    foreach($buf as $iprange => $permrecord)
      $res[$iprange] = $permrecord['parent'];
  }
  return $res;
}

function isOrphan($user, $groupaction = false) {
  global $UserInstanceVars;
  $precord = $UserInstanceVars->getPermHolderRecord($user, $groupaction, false, false, false);
    // a possibly inconsistent record (the third 'false'), since parent might not exist
  return !doesUserExist($precord['parent']);
}

function retainUsersStrictlyBelow($patron, $userList, $groupaction = false, $retain_orphans = false) {
  foreach($userList as $user => $anything) {
    if ((!$retain_orphans || !isOrphan($user, $groupaction)) &&
        !isPatronOf($patron, $user, $groupaction))
      unset($userList[$user]);
  }
  return $userList;
}

function sortUserList($userList, $rootparent) {
  // This code runs in quadratic time in the number of entries in userList.
  ksort($userList, SORT_STRING);
  $res = array();
  $buf[] = $rootparent;
  $processedParent = array_shift($buf);
  while ($processedParent) {
    foreach($userList as $user => $maybeRecord) {
      if (is_array($maybeRecord) && isset($maybeRecord['parent']))
        $userParent = $maybeRecord['parent'];
      else
        $userParent = $maybeRecord;
      if ($userParent == $processedParent) { // if we have found a direct dependent, push it:
        $res[$user] = $maybeRecord;
        $buf[] = $user;
        unset($userList[$user]);
      }
    }
    $processedParent = array_shift($buf);
  }
  return array_merge($res, $userList); // include the remainders "manually" (may be orphans)
}

function getUserListStrictlyBelow($patron) {
  $buf = getUserListEqualOrBelow($patron);
  unset($buf[$patron]);
  return $buf;
}

/* old code
function getUserListEqualOrBelow($patron, $groupaction = false) {
  // Returns a list of all users equal or strictly below $patron 
  // (which is BTW required to be a proper user).
  // [This is implemented far from optimally fast, since we have to
  // look at all user's patron array.]

  global $UA2UserPermDir, $UA2GroupPermDir;

  // first get all users:
  $res = array();
  if ($groupaction) 
    $dirp = @opendir($UA2GroupPermDir);
  else
    $dirp = @opendir($UA2UserPermDir);
  if (!$dirp) return false;
  while (($user=readdir($dirp)) !== false) {
    if ($user[0]=='.') continue;
    $res[$user] = getAllPatrons($user, $groupaction);
  }
  closedir($dirp);

  // then just throw away the ones that we dont want:
  foreach($res as $user => $patronArr) 
    if (!isInArrOfPatrons($patron, $patronArr)) {
      unset($res[$user]); 
    }

  // note that $res[$patron] will be set, since for $user==$patron
  // the corresponding $patronArr contains $patron as well
 
  return $res;
}
*/

//---------------- group membership related ------------------------

//new:
function getAllGroupsOfUser($user, $groupaction = false) {
  // Returns an ass. array xx => 1 containing all groups xx the user is member of.
  // (The construction with the associative array avoids duplicates.)
  global $UA2AllGrpsCallingStack, $UserInstanceVars, $GuestUsergrp, $LoggedInUsergrp;

  // Make sure we dont cycle:
  if (isset($UA2AllGrpsCallingStack[$user]))
    return array();

  $permrecord = $UserInstanceVars->getPermHolderRecord($user, $groupaction);
  if (!$permrecord) return false;
  $res = array();
  foreach($permrecord['perms'] as $granter => $permtable) 
    foreach($permtable as $permentry) {
      if ($permentry{0} == '@') {
        $group = substr($permentry, 1);
        $res[$group] = 1;
        $UA2AllGrpsCallingStack[$group] = 1;
        $res = array_merge($res, getAllGroupsOfUser($group, true));
        unset($UA2AllGrpsCallingStack[$group]);
      } else break; // to conform with the optimization hack used on evaluation
    }
  if (empty($UA2AllGrpsCallingStack) && !$groupaction) {
    $res[$GuestUsergrp] = 1;
    $res[$LoggedInUsergrp] = 1;
  }
  return $res;
}

function IsUserGroupMember($group, $user, $groupaction = false) {
  // Returns true if $user (which means a group if $groupaction is true) is
  // member of group $group.
  $allGroups = getAllGroupsOfUser($user, $groupaction);
  return isset($allGroups[$group]);
}

function IsCurrentUserGroupMember($group) {
  global $UserInstanceVars;
 
  if (! isValidUserString($group, true)) return false; // on bad input silently deny
  if (! $UserInstanceVars->isAuthenticated()) return false;
  return IsUserGroupMember($group, $UserInstanceVars->GetUsername(), false);
}

function IsCurrentUserGroupMemberOld($group) {
  if ($group{0} == '@')
    return IsCurrentUserGroupMember(substr($group, 1));
  return false;
}

/*
//old:
function IsCurrentUserGroupMember($group) {
  global $UserInstanceVars;

  if (! $UserInstanceVars->isAuthenticated()) return false; 
  return IsUserGroupMember($UserInstanceVars->GetUsername(), $group);
}

function IsUserGroupMember($user, $group, $groupaction = false) {
  global $UA2IsUsGrMmCallingStack;

  // Make sure we dont cycle: 
  if (isset($UA2IsUsGrMmCallingStack[$user]))
    return false;

  $UA2IsUsGrMmCallingStack[$user] = 1;
  $res = IsUserGroupMemberWorker($user, $group, $groupaction);
  unset($UA2IsUsGrMmCallingStack[$user]);
  return $res;
  
}

function IsUserGroupMemberWorker($user, $group, $groupaction = false) {
  // Returns true if $user is found to be in group $group, also if
  // this happens via intermediate groups xxx which are members
  // of $group.
  global $UserInstanceVars, $GuestUsergrp, $LoggedInUsergrp;

  if (!$groupaction) { // See #004
    if ($group == $GuestUsergrp) return true;
    if ((strlen($user) > 0) && ($group == $LoggedInUsergrp)) return true;
  }
  
  $permrecord = $UserInstanceVars->getPermHolderRecord($user, $groupaction);

  if ($permrecord) {
    foreach($permrecord['perms'] as $patron => $permtable) {
      foreach($permtable as $permentry) {
        if ($permentry{0} == '@') {
          $usrgroup = substr($permentry, 1);
          if ($usrgroup == $group) return true;
          if (IsUserGroupMember($usrgroup, $group, true)) return true;
        } else break;
      }
    }
  }

  return false;
}
*/

//=============================================================
//======= Perm table related stuff ============================

function isConsistentPermTable(&$permtable, $granter) {
  // Expects an array of alleged perm entries, and returns true if
  // they are all make sense, or if the errors are correctable. (Modifies
  // the argument then.)
  global $UA2LevelToAbbr,
         $UA2AbbrToLevel,
         $UA2PageRelatedLevelAbbr;

  if (!isset($UA2AbbrToLevel)) // make it on the fly if not exists
    foreach($UA2LevelToAbbr as $level => $abbr) 
      $UA2AbbrToLevel[$abbr] = $level;

  foreach($permtable as $permentry) {
    if ($permentry == '*') continue;
    if ($permentry{0} == '#') continue;
    if ($permentry{0} == '@') {
      if (!isValidUserString(substr($permentry, 1), true)) return false;
      if (!isPatronOf($granter, substr($permentry, 1), true)) {
        appendToUA2ErrorLog("Warning: '$granter' granting membership to group '".substr($permentry, 1)
                            ."' though not patron of this group.\n");
      }
    } else {
      if ($permentry{0} == '-') $permentry = substr($permentry, 1);
      $entrylevel = substr($permentry, 0, 2);
      if ($entrylevel == 'xx') continue;
      if (!isset($UA2AbbrToLevel[$entrylevel])) return false; // entry level does not exist
      if (in_array($entrylevel, $UA2PageRelatedLevelAbbr)) {
        if ($permentry{2} != '_') return false;
        $entrypage  = substr($permentry, 3);
        if (!isValidPagenamePattern($entrypage)) return false;  // bad pagename pattern
      }
    }
  }

  return true;
}

//=============================================================
//======= Markup related functions ============================

// conditional variables that can be used in a wiki page
$Conditions['loggedin']="\$GLOBALS['IsUserLoggedIn']";
$Conditions['memberOf']="IsCurrentUserGroupMember(\$condparm)";
$Conditions['member']="IsCurrentUserGroupMemberOld(\$condparm)"; // for backward compatibility
$Conditions['ipMatches'] = 'currClientIpMatchesRange($condparm)';
// The markup of the loginform has to come after UserInstanceVars has been instantiated, and all the current
// writes to SESSION['auth_message'] have taken place, since the GetUserLoginForm() call is already made 
// now, not as usual when assembling the wiki output.
// (The "if" makes sure that the (:loginform:) markup gets only interpreted for the designated login page.)
if ($pagename == $LoginPage) 
  Markup("loginform", "_end", "/\\(:loginform:\\)/", GetUserLoginForm());

//=============================================================
//======= Stuff================================================
function answer($x) { return ($x ? "yes" : "no"); }

//============================================================
//======= Needed by userauth2-admintool.php and -profile.php

function mayCurrAdminActOnPermHolder($action, $admin_action, $tool_username, $groupaction = false) {
  // This is one of the main places where the logic concerning the _granting policy_
  // is concentrated. Here admin actions get mapped to (permtable checkable) permission levels.
  // Expects an existing user $tool_username, unless on create action (where NULL or the new perm holder
  // name is expected).
  global $UserInstanceVars,
         $GuestUsergrp,
         $LoggedInUsergrp;

  $curr_admin = $UserInstanceVars->GetUsername();

  // action is the pmwiki action as communicated via ?action=
  // admin_action is the UA action as communicated via &aa=

  // all of these permission queries are independent of a possible page concerned,
  // thus first argument in hasCurr..() is ''.
  if ($action == 'profile') {
    // the user himself or all his patrons may change his profile if they are entitled:
    if (!isEqualOrPatronOf($curr_admin, $tool_username, $groupaction)) return false;
    if (!HasCurrentUserPermForAction('', 'profile')) return false;
    return true;
  } else
  if ($action == 'pwchange') {
    // only the user himself can change his password:
    if ($tool_username != $curr_admin) return false;
    // but only if he is entitled to do so:
    if (!HasCurrentUserPermForAction('', 'pwchange')) return false;
    return true;
  } else
  if ($action == 'pwset') {
    // setting passwords only for users strictly below the current admin:
    if (!isPatronOf($curr_admin, $tool_username, $groupaction)) return false;
    if (!HasCurrentUserPermForAction('', 'pwset')) return false;
    return true;
  } else
  if ($action == 'admin') {
    // (a) The creation or deletion of guest user or logged in user group is allowed only to 'admin', while
    //     editing is just permitted according to the usual patron principle.
    // (b) Edit and delete actions only strict patrons are entitled to execute (in particular no "self edit" ;)
    //     But this obviously need not apply for creation.
    // (c) creation and deletion of users/groups are subsumed under one level
    // (d) For any action on ip ranges the 'setipperms' level must have been granted

    if (($tool_username == $GuestUsergrp) ||            // (a)
        ($tool_username == $LoggedInUsergrp)) {                   
      if (($admin_action == 'creategroup') || 
          ($admin_action == 'delgroup')) {
        if ($curr_admin != 'admin')
          return false;
      }
    }

    if ($groupaction && (strlen($tool_username) > 0) && isIpRange($tool_username) &&
        !HasCurrentUserPerm('', 'setipperms')) return false; // (d)

    if ($admin_action == 'createuser') {
      return HasCurrentUserPerm('', 'createuserperms'); // (c)
    } else
    if ($admin_action == 'creategroup') {
      return HasCurrentUserPerm('', 'createuserperms'); // (c)
    } else
    if ($admin_action == 'deluser') {
      if (!isPatronOf($curr_admin, $tool_username, $groupaction)) return false; // (b)
      return HasCurrentUserPerm('', 'createuserperms'); // (c)
    } else
    if ($admin_action == 'delgroup') {
      if (!isPatronOf($curr_admin, $tool_username, $groupaction)) return false; // (b)
      return HasCurrentUserPerm('', 'createuserperms'); // (c)
    } else
    if ($admin_action == 'edituser') {
      if (!isPatronOf($curr_admin, $tool_username, $groupaction)) return false; // (b)
      return HasCurrentUserPerm('', 'edituserperms');
    } else
    if ($admin_action == 'editgroup') {
      if (!isPatronOf($curr_admin, $tool_username, $groupaction)) return false; // (b)(a)
      return HasCurrentUserPerm('', 'edituserperms');
    } else
    if ($admin_action == 'setipperms') {
      if (!isPatronOf($curr_admin, $tool_username, $groupaction)) return false; // (b)
      return HasCurrentUserPerm('', 'setipperms');
    }
  }

  appendToUA2ErrorLog("Warning: Unknown action/admin_action pair: '$action/$admin_action'. Refused.\n");
  return false;
}

function containsChar($needleChr, $haystack) {
  return strpos($haystack, $needleChr) !== false;
}

function isSecure($str) {
  if (containsChar('<', $str)) return false;
  if (containsChar('>', $str)) return false;
  if (containsChar('&', $str)) return false;
  return true;
}

//================================================================
//======== Operations executed effectively once after install ====

mkdirp($UA2UserPermDir); // making sure that dirs exist
mkdirp($UA2GroupPermDir);
mkdirp($UA2ProfileDir);
mkdirp($UA2IpRangesDir);
if (!loadUserProfile('admin')) { // making sure that admin profile exist
  $prof = newEmptyProfile();     // (will have an empty password)
  saveUserProfile('admin', $prof);
}
if (!loadUserPermRecord($GuestUsergrp, true)) { 
  $rec = newEmptyPermRecord('admin', true);
  $mynewpermtable = array();
  $mynewpermtable[] = 'rd_*.*'; // on initial start, have a read-all for guest users
  $rec['perms']['admin'] = $mynewpermtable;
  saveUserPermRecord($GuestUsergrp, $rec, true);
}

//================================================================
//======== admin tool and passwd change inclusions ===============

if ($action == 'admin') // conditionally include scripts if needed
  require_once("userauth2/userauth2-admintool.php");
if ($action == 'pwchange') {
  require_once("userauth2/userauth2-pwchange.php");
}

